# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for signal filter plugin
    This filter executes one or more signal processing transformations on input data
"""

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os

import logging
from foglamp.common import logger
from foglamp.plugins.common import utils
from foglamp.common.plugin_helpers import HandleMap, PluginHandle

import filter_ingest


_LOGGER = logger.setup(__name__, level=logging.DEBUG)

_module_dir = os.path.dirname(__file__)
_handle_map = HandleMap(__name__)

FOGLAMP_ROOT = os.getenv("FOGLAMP_ROOT", "/usr/local/foglamp")
FOGLAMP_DATA = os.getenv("FOGLAMP_DATA", os.path.join(FOGLAMP_ROOT, "data"))

DEFAULT_N_FFT = 8192
DEFAULT_SAMPLING_RATE = 8000

class SignalProHandle(PluginHandle):

    # mandatory parameters for PluginHandle which has an rpc back-end
    RPC_SERVER_NAME='sigfns_server'          # name of RPC server module
    RPC_CONFIG_MEMBERS=['filter_list', 'data_column', 'output_dir', 'result_type',
                        'sampling_rate', 'n_fft'
    ]   # config members to send to rpc for its config

    def __init__(self, config, cbinfo):
        super().__init__(__name__, cbinfo)  # set up logger, callback

        # slots -- values will be set by config
        self.plugin = None
        self.enable = None
        self.input_asset = None
        self.output_asset = None
        self.data_column = None
        self.forward_data = None
        self.sampling_rate = None
        self.n_fft = None
        self.sp_multi_filter = None

        self.rpc = None
        self.input = []
        self.input_type = "ts" # xxx - may want to derive from data or configuration

        self.reconfigure(config)  # first time through, creates and configures server

    def reconfigure(self, new_config):
        """ reconfigure - make the current plugin state consistent with the new configuration """

        # our server state is pretty simple, just tell the plugin to re-form its filters
        self.config_update(new_config)
        self.rpc_setup(self._rpc_config(), _module_dir)

    def config_update(self, update):
        """ config_update - save config values as members in this handle """
        super().config_update(update)       # update the values in the handle

        self.FRAME_LEN = self.sampling_rate
        self.output_dir = os.path.join(FOGLAMP_DATA, self.output_asset)
        os.makedirs(self.output_dir, exist_ok=True)

        if self.sp_multi_filter == '':
            raise ValueError("must specify spMultiFilter with comma-separated filter list")
        _filter_list = self.sp_multi_filter.split(',')

        self.result_type = self.input_type
        for f in _filter_list:
            # look for transitions from time series to frequency
            if f == 'fft':
                self.result_type = "freq"
            # xxx - if we add rfft, need to switch back to time series
        
        self.filter_list = _filter_list

    def ingest(self, data):
        """ ingest - main ingest entry point, send data frames to signal processing server """

        if type(data) is dict:
            data = [data]

        _processed_results = []
        _forward_assets = []
        
        for d in data:
            if not self.enable or \
               d['asset'] != self.input_asset:
                _forward_assets.append(d)
                continue

            if self.forward_data:
                _forward_assets.append(d)

            _readings = d['readings']
            _names = _readings.get('name', None)
            if _names:
                # readings are already frames of data, let server process them
                if type(_names) is str:
                    _names = [_names]
                _dir = _readings['dir']
                
                _processed_results = self.rpc.process_frames(_dir, _names)
                
            else:
                # much less efficient -- readings are raw data, pick out the data column and
                # send it to the server, get back a data frame file
                if self.data_column == '': raise ValueError("data column must be specified to be one of {}".format(list(_readings.keys())))
                
                # only send data col to filter. filter(data_col) is all that goes forward
                # manage whether we're getting one entry per reading or multiple readings
                if type(_readings) is dict:
                    self.input.append(_readings[self.data_column])
                else:
                    r_data = [r[self.data_column] for r in _readings]
                    self.input.extend(r_data)

                # chop it up into frames
                _chunks = []
                while len(self.input) >= self.FRAME_LEN:
                    _chunks.append(self.input[:self.FRAME_LEN])
                    self.input = self.input[self.FRAME_LEN:]

                for chunk in _chunks:
                    # call to the server process to transform the frame using its filters
                    _processed_results.append(self.rpc.process_data(chunk))

        # forward any existing or unrelated assets
        if len(_forward_assets) > 0:
            filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, _forward_assets)

        # return the asset with our processed results -- processed results are a list of frame files and a dir
        if len(_processed_results) > 0:
            # self.logger.info("output[{}]: {}".format(self.output_asset, str(_processed_results)[:200]))
            _result = [{
                'asset': self.output_asset,
                'readings': {
                    'datafmt': 'FRAME',
                    'dir': self.output_dir,
                    'name': _processed_results,
                    'sampling_rate': self.sampling_rate
                }
            }]

            filter_ingest.filter_ingest_callback(self.callback, self.ingest_ref, _result)

    def plugin_shutdown(self):
        if self.rpc:
            self.rpc.plugin_shutdown()
            self.rpc = None
    
#
#
### PLUGIN API ###
#
#

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Signal processing filter plugin',
        'type': 'string',
        'default': 'signal',
        'readonly': 'true'
    },
    'enable': {
        'description': 'enable/disable this filter',
        'type': 'boolean',
        'default': 'true',
        'displayName': 'Asset name of time series data',
        'order': "1"
    },
    'inputAsset': {
        'description': 'Input asset name',
        'type': 'string',
        'default': 'timeseries',
        'displayName': 'Asset name of time series data',
        'order': "2"
    },
    'outputAsset': {
        'description': 'Name of asset where filtered signal is written',
        'type': 'string',
        'default': 'signalProcessedOutput',
        'displayName': 'Output data asset',
        'order': "3"
    },
    'dataColumn': {
        'description': 'Name of the key within readings with data we are filtering',
        'type': 'string',
        'default': '',
        'displayName': 'Data column key',
        'order': "4"
    },
    'forwardData': {
        'description': 'Forward input data through filter',
        'type': 'boolean',
        'default': 'false',
        'displayName': 'Forward input data',
        'order': "5"
    },
    'samplingRate': {
        'description': 'Sampling rate of acceleration data',
        'type': 'integer',
        'default': str(DEFAULT_SAMPLING_RATE),
        'displayName': 'Sampling rate (samples per sec)',
        'order': "6"
    },
    'nFft': {
        'description': 'Default number of FFT lines for fft operations',
        'type': 'integer',
        'default': str(DEFAULT_N_FFT),
        'displayName': 'Number of FFT lines',
        'order': "7"
    },
    'spMultiFilter': {
        'description': 'Comma-separated list of filter names and arguments',
        'type': 'string',
        'default': '',
        'displayName': 'One or more filter specs (comma separated)',
        'order': "8"
    },
}

def plugin_info():
    """ Returns information about the plugin
    Args:
    Returns:
        dict: plugin information
    Raises:
    """
    _LOGGER.debug("plugin_info called")
    return {
        'name': 'sigfns',
        'version': '1.9.2',
        'mode': "none",
        'type': 'filter',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }

def plugin_init(config, ingest_ref, callback):
    """ Initialise the plugin
    Args:
        config: JSON configuration document for the Filter plugin configuration category
        ingest_ref:
        callback:
    Returns:
        data: JSON object to be used in future calls to the plugin
    Raises:
    """
    _handle = SignalProHandle(config, (ingest_ref, callback))
    _h = _handle_map.new_handle(_handle, config)
    
    return _h

def plugin_ingest(h, data):
    """ Absorb run signal processing algorithms on time series data "data"
    Args:
        handle: handle returned by the plugin initialisation call
        data: readings data
    """

    _handle = _handle_map.get_handle(h)

    _handle.ingest(data)

def plugin_reconfigure(h, new_config):
    """ Reconfigures the plugin

    Args:
        h: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    """
    _handle = _handle_map.get_handle(h)

    _handle.reconfigure(new_config)

    h['config'].update(new_config)

    return h

def plugin_shutdown(h):
    """ Shutdowns the plugin doing required cleanup.
    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        plugin shutdown
    """
    _LOGGER.info('plugin_shutdown called')

    _handle = _handle_map.get_handle(h)
    _handle.plugin_shutdown()  # shutdown any RPC objects
    
    _handle_map.del_handle(h)  # remove our mapping
