# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Module for signal filter plugin
    This filter executes one or more signal processing transformations on input data
"""

__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os

import math

import numpy as np
import pandas as pd
from scipy import signal

from datetime import datetime

import logging
from foglamp.common import logger
from foglamp.common.plugin_helpers import PluginRPCServer

import sp
import sp_vib

FORMAT_TS="%Y-%m-%d %H:%M:%S.%f"


class SignalFnsServer(PluginRPCServer):
    """ process data frames applying signal processing filters to them """

    FILTERS=['highpass', 'lowpass', 'bandpass', 'stop', 'notch', 'resample', 'hilbert', 'denoise',
               'autocorrelate', 'detrend', 'normalize', 'wiener', 'savgol', 'fft', 'comb', 'denoise_freq']

    def __init__(self, service_name):
        super().__init__(service_name)  # set up logging, iprpc server

        # slots -- values will be assigned by config_update
        self.sampling_rate = None
        self.sp_multi_filter = None

        self.logger.setLevel(logging.DEBUG)

    def plugin_init(self, config):
        # plugin_init -- apply the server params to the current handle

        # signal server only manages the filter_list parameter
        # apply config params to current handle
        self.config_update(config)
        
        self.nyq = self.sampling_rate / 2

        self.logger.info("filter list: {}".format(self.filter_list))

        # convert the multi-filter into a tuple of (filter-fn, (arg-tuple))
        def _filter_op(t):
            if len(t) > 2:
                raise ValueError("invalid filter, too many arguments {}".format(t))
            _op = t[0]
            if _op not in self.FILTERS:
                raise ValueError("Unknonwn filter operation {}".format(_op))
            if not hasattr(self, _op):
                raise ValueError("Unimplemented filter operation {}".format(_op))
                
            try:
                _fop = getattr(self, _op)
                if len(t) == 3:
                    _f = (_fop, (float(t[1]), float(t[2])))
                elif len(t) == 2:
                    _f = (_fop, (float(t[1]),))
                else:
                    _f = (_fop, ())
            except:
                raise ValueError("could not parse filter operation {}".format(t))
            return _f

        # create the functional filter operation list, including optional parameters
        self.filter_list = list(map(lambda x: _filter_op(x.split(':')), self.filter_list))


    def _process(self, data):
        """ process frame - receive one frame of one column of data; 
            convert using sp filters and return the converted result """

        # execute the filter(s) sequentially
        for f in self.filter_list:
            data = f[0](data, *f[1])

        return data

    def process_data(self, data):
        """ process_data - take an array of raw data, process, and write out and return frame file """
        _data = np.asarray(data)
        _data = self._process(_data)
        
        _fname = self._flush(_data)
        return [_fname]

    def process_frames(self, f_dir, fnames):
        """ process_frame - take a frame file of data, process, write out and return new frame file """
        _out_names = []

        if self.data_column == '':
            raise ValueError("must specify dataColumn to extract frame data")

        for fname in fnames:
            if not fname.endswith(".tspkl"):
                raise ValueError("{}: invalid frame type".format(_in_path))
            
            _in_path = os.path.join(f_dir, fname)
            _frame = pd.read_pickle(_in_path)

            if self.data_column not in list(_frame.columns):
                raise ValueError("dataColumn '{}' not found in dataframe columns {}".format(self.data_column, list(_frame.columns)))

            _data = _frame[self.data_column].to_numpy()
            _data = self._process(_data)

            # make us a new frame
            _out_names.append(self._flush(_data))
            
        return _out_names

    def _flush(self, data):
        """ flush - create a pickle file with this frame of transformed data in it """
        def _gen_filename(pfx, ts, *args, suff="tspkl"):
            """ gen_filename - generate a new, unique file name """
            _seq = 0
            while True:
                _fname = "{}_{}_{}".format(pfx, ts.strftime(FORMAT_TS), _seq)
                for a in args: _fname += "_{}".format(a)
                _fname += "."+suff
                if not os.path.exists(os.path.join(self.output_dir, _fname)):
                    break
                _seq += 1
            return _fname

        if self.result_type == "freq":
            # end result is to create an fft file
            _fname = _gen_filename("FFT", datetime.utcnow(), len(data), suff="fftpkl")
            _freq, _amp = data   # data is a tuple
            data = {"freq": _freq, "amp": _amp}
        else:
            # timeseries is the default
            _fname = _gen_filename("TS", datetime.utcnow(), len(data))
            # assign an arbitrary data column label
            data = {"y": data}
            
        _out_path = os.path.join(self.output_dir, _fname)

        _frame = pd.DataFrame(data)
        _frame.to_pickle(_out_path)

        return _fname
        

    def plugin_shutdown(self):
        pass
    
    #
    ###  filter operations  ###
    #
    def fft(self, ts) -> tuple:  # tuple of freqs, amps
        return sp.fft(ts, fs=self.sampling_rate, N_fft=self.n_fft)

    def normalize(self, ser):
        """ normalize - normalize amplitude to zero """
        if type(ser) is tuple:
            # frequency 
            _freq, _amp = ser
            _max = np.max(_amp)
            if _max == 0:
                return ser
            return (_freq, _amp / _max)

        # time
        _max = np.max(ser)
        return ser / _max
    
    def denoise_freq(self, ser):
        """ denoise_freq - get rid of chatter that will distract dnn's """
        NOISE_LIMIT = .001     # assume we're normalized, so we erase when less than 2 orders of magnitude
        if type(ser) is tuple:
            # frequency
            _freq, _amp = ser
            _idx = np.argsort(_amp)
            _amp = np.where(_amp > NOISE_LIMIT, _amp, 0)
            return (_freq, _amp)

        # time
        return ser
    
    def resample(self, ts, nsample):
        return sp.resample_y_to(ts, nsample)
    
    def highpass(self, ts, cutoff):
        cutoff /= self.nyq
        return sp.bandwidth_filter_high(ts, cutoff)

    def lowpass(self, ts, cutoff):
        cutoff /= self.nyq
        return sp.bandwidth_filter_low(ts, cutoff)

    def bandpass(self, ts, lower, upper):
        lower /= self.nyq
        upper /= self.nyq
        return sp.bandwidth_filter_bandpass(ts, [lower, upper])

    def bandstop(self, ts, lower, upper):
        lower /= self.nyq
        upper /= self.nyq
        return sp.bandwidth_filter_bandstop(ts, [lower, upper])

    def notch(self, ts, notch_f):
        w0 = notch_f / self.nyq 
        if w0 >= 1 or w0 <= 0:
            raise ValueError("error: filter values must be (0,{})".format(self.nyq))
        Q=30
        b, a = signal.iirnotch(w0, Q)
        return signal.filtfilt(b, a, ts, axis=0)
        
    def comb(self, ts, notch_f):
        self.logger.info("comb: f={} shape={}".format(notch_f, len(ts)))
        w0 = notch_f / self.nyq 
        if w0 >= 1 or w0 <= 0:
            raise ValueError("error: filter values must be (0,{})".format(self.nyq))
        Q=30
        b, a = signal.iircomb(w0, Q)
        return signal.filtfilt(b, a, ts)

    def hilbert(self, ts):
        return sp.hilbert(ts)
    
    def denoise(self, ts):
        return sp_vib.denoise(ts)
    
    def autocorrelate(self, ts):
        return sp_vib.autocorrelate(ts, self.sampling_rate)

    def detrend(self, ts):
        return signal.detrend(ts)
    
    def wiener(self, ts):
        return signal.wiener(ts)

    def savgol(self, ts):
        WIN_LEN = 101
        POLY_ORDER = 3
        return signal.savgol_filter(ts, WIN_LEN, POLY_ORDER)
    

server = SignalFnsServer('sp-fns')
server.serve()
