# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" standard signal processing functions """

import numpy as np

import scipy.signal as signal
import scipy.fftpack as fftpack


def fft(data, fs, N_fft):
    
    def _get_fft_bin_values(y_values, T, N, fs, N_fft):
        mean_removed = np.ones_like(y_values)*np.mean(y_values)
        y_values = y_values - mean_removed

        window = np.hanning(y_values.shape[0])
        yf = fftpack.fft(y_values*window, n=N_fft)

        freqs = np.arange(0, fs//2, fs/N_fft)
        amps = 2.0/N_fft * np.abs(yf[0:N_fft//2])

        return freqs, amps

    N = data.shape[0]   # num of samples
    T = 1./fs
    
    freqs, amps = _get_fft_bin_values(data, T, N, fs, N_fft)
    return freqs, amps


def filter_signal(data, sampling_rate, f_lower, f_upper):
    """ Function to filter signal
    Args:
        data: Signal data
        sampling_rate: Sampling rate at which vibration data was collected
        f_lower: Lower bound for filter
        f_upper: Upper bound for filter
    Returns:
        Returns filtered signal
    """

    nyq = sampling_rate / 2.0
    if (f_lower == 0) and (f_upper == sampling_rate / 2):
        pass
    elif f_lower == 0:
        data = bandwidth_filter_low(data, f_upper / nyq)
    elif f_upper == sampling_rate / 2:
        data = bandwidth_filter_high(data, f_lower / nyq)
    else:
        data = bandwidth_filter_bandpass(data, [f_lower / nyq, f_upper / nyq])

    return data

def bandwidth_filter_high(data, cutoff, order=10):
    _sos = signal.butter(order, cutoff, btype='high', output='sos')
    return signal.sosfiltfilt(_sos, data)

def bandwidth_filter_low(data, cutoff, order=10):
    _sos = signal.butter(order, cutoff, btype='low', output='sos')
    return signal.sosfiltfilt(_sos, data)

def bandwidth_filter_bandpass(data, low_high_cutoff, order=10):
    _sos = signal.butter(order, low_high_cutoff, btype='bandpass', output='sos')
    return signal.sosfiltfilt(_sos, data)

def bandwidth_filter_bandstop(data, low_high_cutoff, order=10):
    _sos = signal.butter(order, low_high_cutoff, btype='bandstop', output='sos')
    return signal.sosfiltfilt(_sos, data)

def hilbert(data):
    # demodulate and extract envelope
    analytic_signal = signal.hilbert(data)
    return np.abs(analytic_signal)


def resample_xy_to(x, y, npoints):
        # create a sampled subset of the entire dataset
        if len(y) <= npoints:
            # no resample necessary
            return x, y

        _y_samp = resample_y_to(y, npoints)

        # update x to match y
        _x_samp = np.linspace(x[0], x[-1], len(_y_samp), endpoint=False)

        return _y_samp, _x_samp
                          

RESAMPLE_CHUNK_SIZE=8192
def resample_y_to(y, npoints):
    _org_len = len(y)
    if _org_len <= npoints:
        # no resample necessary
        return y

    npoints = min(RESAMPLE_CHUNK_SIZE*2, npoints)  # floor

    # resample chunk-wise; otherwise the signal resampler acts like too much of a lowpass filter
    _resample_ratio = npoints / _org_len
            
    _CHUNK = RESAMPLE_CHUNK_SIZE # convert each chunk of this length...
    _sampled_chunk = max(_CHUNK//16, int(_CHUNK * _resample_ratio)) # ...to this

    _rem = _org_len % _CHUNK
    _n_chunk = _org_len // _CHUNK

    _new_len = _sampled_chunk*_n_chunk
    _y_samp = np.empty(_new_len)

    _y_idx = _s_idx = 0

    # resample piece-wise in big chunks
    for i in range(0, _n_chunk-1):
        # resample each segment
        _y_samp[_s_idx:_s_idx+_sampled_chunk] = signal.resample(y[_y_idx:_y_idx+_CHUNK], _sampled_chunk)
        _y_idx += _CHUNK
        _s_idx += _sampled_chunk

    # pick up the last 1.x chunks as one piece
    if _new_len > _s_idx:
        _y_samp[_s_idx:_new_len] = signal.resample(y[_y_idx:_org_len], _sampled_chunk)

    return _y_samp
